/* Main processing code for Services.
 *
 * Services is copyright (c) 1996-1997 Andy Church.
 *     E-mail: <achurch@dragonfire.net>   IRC: Alcan@EsperNet
 * This program is free but copyrighted software; see the file COPYING for
 * details.
 */

#include "services.h"
#include "version.h"

/*************************************************************************/

/* Use ignore code? */

int allow_ignore = 1;


/* People to ignore (hashed by first character of nick). */

IgnoreData *ignore[256];


/* add_ignore: Add someone to the ignorance list for the next `delta'
 *             seconds.
 */

void add_ignore(const char *nick, time_t delta)
{
    IgnoreData *ign;
    char who[NICKMAX];
    time_t now = time(NULL);
    IgnoreData **whichlist = &ignore[tolower(nick[0])];

    strscpy(who, nick, NICKMAX);
    for (ign = *whichlist; ign; ign = ign->next) {
	if (stricmp(ign->who, who) == 0)
	    break;
    }
    if (ign) {
	if (ign->time > now)
	    ign->time += delta;
	else
	    ign->time = now + delta;
    } else {
	ign = smalloc(sizeof(*ign));
	strcpy(ign->who, who);
	ign->time = now + delta;
	ign->next = *whichlist;
	*whichlist = ign;
    }
}


/* get_ignore: Retrieve an ignorance record for a nick.  If the nick isn't
 *             being ignored, return NULL and flush the record from the
 *             in-core list if it exists (i.e. ignore timed out).
 */

IgnoreData *get_ignore(const char *nick)
{
    IgnoreData *ign, *prev;
    time_t now = time(NULL);
    IgnoreData **whichlist = &ignore[tolower(nick[0])];

    for (ign = *whichlist, prev = NULL; ign; prev = ign, ign = ign->next) {
	if (stricmp(ign->who, nick) == 0)
	    break;
    }
    if (ign && ign->time <= now) {
	if (prev)
	    prev->next = ign->next;
	else
	    *whichlist = ign->next;
	free(ign);
	ign = NULL;
    }
    return ign;
}

/*************************************************************************/

/* split_buf:  Split a buffer into arguments and store the arguments in an
 *             argument vector pointed to by argv (which will be malloc'd
 *             as necessary); return the argument count.  All argument
 *             values will also be malloc'd.  If colon_special is non-zero,
 *             then treat a parameter with a leading ':' as the last
 *             parameter of the line, per the IRC RFC.  Destroys the buffer
 *             by side effect.
 */

int split_buf(char *buf, char ***argv, int colon_special)
{
    int argvsize = 8;
    int argc;
    char *s, *t;

    if (!(*argv = malloc(sizeof(char *) * argvsize)))
	return -1;
    argc = 0;
    while (*buf) {
	if (argc == argvsize) {
	    argvsize += 8;
	    if (!(*argv = realloc(*argv, sizeof(char *) * argvsize)))
		return -1;
	}
	if (*buf == ':') {
	    (*argv)[argc++] = buf+1;
	    buf = "";
	} else {
	    s = strpbrk(buf, " ");
	    if (s) {
		*s++ = 0;
		while (isspace(*s))
		    s++;
	    } else {
		s = buf + strlen(buf);
	    }
	    (*argv)[argc++] = buf;
	    buf = s;
	}
    }
    return argc;
}

/*************************************************************************/

/* process:  Main processing routine.  Takes the string in inbuf (global
 *           variable) and does something appropriate with it. */

void process()
{
    char source[64];
    char cmd[64];
    char buf[512];		/* Longest legal IRC command line */
    char *s;
    int ac;			/* Parameters for the command */
    char **av;
    time_t starttime, stoptime;	/* When processing started and finished */


    /* If debugging, log the buffer. */
    if (debug)
	log("debug: Received: %s", inbuf);

    /* First make a copy of the buffer so we have the original in case we
     * crash - in that case, we want to know what we crashed on. */
    strscpy(buf, inbuf, sizeof(buf));

    /* Split the buffer into pieces. */
    if (*buf == ':') {
	s = strpbrk(buf, " ");
	if (!s)
	    return;
	*s = 0;
	while (isspace(*++s))
	    ;
	strscpy(source, buf+1, sizeof(source));
	strcpy(buf, s);
    } else {
	*source = 0;
    }
    if (!*buf)
	return;
    s = strpbrk(buf, " ");
    if (s) {
	*s = 0;
	while (isspace(*++s))
	    ;
    } else
	s = buf + strlen(buf);
    strscpy(cmd, buf, sizeof(cmd));
    ac = split_buf(s, &av, 1);



    /* Do something with the command. */

    if (stricmp(cmd, "PING") == 0) {

	send_cmd(SERVER_NAME, "PONG %s %s", ac>1 ? av[1] : server_name, av[0]);

    } else if (stricmp(cmd, "436") == 0) {  /* Nick collision caused by us */

#if !defined(SKELETON) && !defined(READONLY)
	introduce_user(av[0]);
#endif

    } else if (stricmp(cmd, "AWAY") == 0) {

#ifndef SKELETON
	if (ac == 0 || *av[0] == 0)	/* un-away */
	    check_memos(source);
#endif

    } else if (stricmp(cmd, "GLOBOPS") == 0
	    || stricmp(cmd, "GNOTICE") == 0
	    || stricmp(cmd, "GOPER"  ) == 0
	    || stricmp(cmd, "WALLOPS") == 0) {

	/* Do nothing */

    } else if (stricmp(cmd, "JOIN") == 0) {

	if (ac != 1)
	    return;
	do_join(source, ac, av);

    } else if (stricmp(cmd, "KICK") == 0) {

	if (ac != 3)
	    return;
	do_kick(source, ac, av);

    } else if (stricmp(cmd, "KILL") == 0) {

	if (ac != 2)
	    return;
	do_kill(source, ac, av);

    } else if (stricmp(cmd, "MODE") == 0) {

	if (*av[0] == '#' || *av[0] == '&') {
	    if (ac < 2)
		return;
	    do_cmode(source, ac, av);
	} else {
	    if (ac != 2)
		return;
	    do_umode(source, ac, av);
	}

    } else if (stricmp(cmd, "MOTD") == 0) {

	/* MOTD command */

	send_cmd(server_name, "375 %s :- %s Message of the Day", source, server_name);
	send_cmd(server_name, "372 %s :- Welcome to EsperNet! <http://www.esper.net>", source);
	send_cmd(server_name, "372 %s :- EsperNet's Services include:", source);
	send_cmd(server_name, "372 %s :- \2ChanServ\2, for channel registration: \2/msg ChanServ HELP\2", source);
	send_cmd(server_name, "372 %s :- \2NickServ\2, for nickname registration: \2/msg NickServ HELP\2", source);
	send_cmd(server_name, "372 %s :- \2MemoServ\2, for sending memos to registered users: \2/msg MemoServ HELP\2", source);
	send_cmd(server_name, "372 %s :- \2HelpServ\2, for EsperNet info: \2/msg HelpServ HELP\2", source);
	send_cmd(server_name, "372 %s :- \2IrcIIHelp\2, for help with the ircII client: \2/help \37topic\37\2", source);
	send_cmd(server_name, "372 %s :- ", source); 
	send_cmd(server_name, "372 %s :- The following can recover nick/channel passwords:", source);
	send_cmd(server_name, "372 %s :- Alcan, AndrewK, Darien, Delirium, IJ", source);
	send_cmd(server_name, "376 %s :End of /MOTD command.", source);

    } else if (stricmp(cmd, "NICK") == 0) {

	if ((!*source && ac != 7) || (*source && ac != 2))
	    return;
	do_nick(source, ac, av);

    } else if (stricmp(cmd, "NOTICE") == 0) {

	/* Do nothing */

    } else if (stricmp(cmd, "PART") == 0) {

	if (ac < 1 || ac > 2)
	    return;
	do_part(source, ac, av);

    } else if (stricmp(cmd, "PASS") == 0) {

	/* Do nothing - we assume we're not being fooled */

    } else if (stricmp(cmd, "PRIVMSG") == 0) {

	if (ac != 2)
	    return;

	/* Check if we should ignore.  Operators always get through. */
	if (allow_ignore && !is_oper(source)) {
	    IgnoreData *ign = get_ignore(source);
	    if (ign && ign->time > time(NULL)) {
		log("Ignored message from %s: \"%s\"", source, inbuf);
		return;
	    }
	}

	starttime = time(NULL);

	if (stricmp(av[0], s_OperServ) == 0) {
	    if (is_oper(source))
		operserv(source, av[1]);
	    else
		notice(s_OperServ, source, "Access denied.");
	} else if (stricmp(av[0], s_NickServ) == 0) {
	    nickserv(source, av[1]);
	} else if (stricmp(av[0], s_ChanServ) == 0) {
	    chanserv(source, av[1]);
	} else if (stricmp(av[0], s_MemoServ) == 0) {
	    memoserv(source, av[1]);
	} else if (stricmp(av[0], s_HelpServ) == 0) {
	    helpserv(s_HelpServ, source, av[1]);
	} else if (stricmp(av[0], "ircIIhelp") == 0) {
	    char *s = smalloc(strlen(av[1]) + 7);
	    sprintf(s, "ircII %s", av[1]);
	    helpserv("IrcIIHelp", source, s);
	    free(s);
	}

	/*Add to ignore list if the command took a significant amount of time.*/
	if (allow_ignore) {
	    stoptime = time(NULL);
	    if (stoptime > starttime && *source && !strchr(source, '.'))
		add_ignore(source, stoptime-starttime);
	}

    } else if (stricmp(cmd, "QUIT") == 0) {

	if (ac != 1)
	    return;
	do_quit(source, ac, av);

    } else if (stricmp(cmd, "SERVER") == 0 || stricmp(cmd, "SQUIT") == 0) {

	/* Do nothing.  Should we eventually do something? */

    } else if (stricmp(cmd, "TOPIC") == 0) {

	if (ac != 4)
	    return;
	do_topic(source, ac, av);

    } else if (stricmp(cmd, "USER") == 0) {

	/* Do nothing - we get everything we need from the NICK command. */

    } else if (stricmp(cmd, "VERSION") == 0) {

	if (source)
	    send_cmd(server_name, "351 %s %s %s :-- %s",
			source, version_number, server_name, version_build);

    } else {

	log("unknown message from server (%s)", inbuf);

    }

}
